/* See LICENSE file for copyright and license details. */
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdarg.h>
#include <errno.h>
#include "tree.h"
#include "spinlock.h"

/* Memory alignment for the allocation functions */
enum { ALIGN = 4 * sizeof(size_t) };

static void writelog(int fd, const char *fmt, ...);
static void dumpstats(void);

struct lemon_stats {
	/* # of sbrk calls */
	unsigned long nr_sbrk;
	/* # of mmap calls */
	unsigned long nr_mmap;
	/* # of malloc calls */
	unsigned long nr_malloc;
	/* # of realloc calls */
	unsigned long nr_realloc;
	/* # of free calls */
	unsigned long nr_free;
	/* # of nodes in the allocation tree */
	unsigned long nr_at_nodes;
	/* # of nodes in the free tree */
	unsigned long nr_ft_nodes;
	/* # of new allocations */
	unsigned long nr_alloc_new;
	/* # of allocations satisfied from the free tree */
	unsigned long nr_alloc_free;
	/* # of shrinked reallocs */
	unsigned long nr_shrink_realloc;
	/* # of invalid frees */
	unsigned long nr_invalid_free;
};

static bool init;
static struct lemon_stats st;
static bool strict = false;

struct node {
	void *buf;
	size_t sz;
	RB_ENTRY(node) entry;
};

RB_HEAD(free_tree, node) ft = RB_INITIALIZER(&ft);
RB_HEAD(alloc_tree, node) at = RB_INITIALIZER(&at);
static spinlock_t rblock;

static int ft_cmp(struct node *a, struct node *b);
RB_PROTOTYPE(free_tree, node, entry, ft_cmp)

static int at_cmp(struct node *a, struct node *b);
RB_PROTOTYPE(alloc_tree, node, entry, at_cmp)

RB_GENERATE(free_tree, node, entry, ft_cmp)
/* These are ordered by `sz' */
static int
ft_cmp(struct node *a, struct node *b)
{
	if (a->sz < b->sz)
		return -1;
	if (a->sz > b->sz)
		return 1;
	return 0;
}

RB_GENERATE(alloc_tree, node, entry, at_cmp)
/* These are ordered by address */
static int
at_cmp(struct node *a, struct node *b)
{
	if (a->buf < b->buf)
		return -1;
	if (a->buf > b->buf)
		return 1;
	return 0;
}

static inline void *
align_pointer(size_t align, void *p)
{
	return (void *)(((uintptr_t)p + align) & ~(align - 1));
}

static inline void *
alloc_object(size_t sz)
{
	void *base, *p;

	base = sbrk(sz + ALIGN);
	if (base == (void *)-1)
		return NULL;
	p = base;
	st.nr_sbrk++;
	return align_pointer(ALIGN, p);
}

static inline void *
mmap_aligned(size_t align, size_t sz)
{
	void *p;

	/* align should be a power of two */
	if ((align - 1) & align)
		return NULL;
	p = mmap(0, sz + align, PROT_READ | PROT_WRITE,
		 MAP_PRIVATE | MAP_ANON, -1, 0);
	if (p == MAP_FAILED)
		return NULL;
	st.nr_mmap++;
	return align_pointer(align, p);
}

static int
lemon_init(void)
{
	char *p;

	if (init)
		return 0;
	p = getenv("LEMONCAKE_DEBUG");
	if (p)
		atexit(dumpstats);
	init = true;
	return 0;
}

void *
malloc(size_t sz)
{
	struct node n, *an, *res;
	void *p;

	if (lemon_init())
		return NULL;
	if (!sz)
		return NULL;
	if (sz > SIZE_MAX / 2)
		return NULL;
	lock(&rblock);
	/* Lookup in the free tree for a block greater
	 * than or equal to `sz' bytes */
	n.sz = sz;
	res = RB_NFIND(free_tree, &ft, &n);
	if (!res) {
		/* No available free block, create a new block
		 * and add it to the alloc tree */
		an = alloc_object(sizeof(*an));
		if (!an) {
			unlock(&rblock);
			return NULL;
		}
		p = mmap_aligned(ALIGN, sz);
		if (!p) {
			unlock(&rblock);
			return NULL;
		}
		an->buf = p;
		an->sz = sz;
		RB_INSERT(alloc_tree, &at, an);
		unlock(&rblock);
		st.nr_malloc++;
		st.nr_alloc_new++;
		return an->buf;
	}
	an = RB_REMOVE(free_tree, &ft, res);
	RB_INSERT(alloc_tree, &at, an);
	unlock(&rblock);
	st.nr_malloc++;
	st.nr_alloc_free++;
	return an->buf;
}

static inline void
align_node(struct node *n)
{
	void *p = n->buf;

	if (n->sz < ALIGN)
		return;
	n->buf = align_pointer(ALIGN, n->buf);
	n->sz = n->sz - ((uintptr_t)n->buf - (uintptr_t)p);
}

void *
realloc(void *oldp, size_t sz)
{
	struct node n, *res;
	struct node *oldan, *newan;
	struct node *fn, *newfn;

	if (lemon_init())
		return NULL;
	if (!oldp)
		return malloc(sz);
	if (!sz) {
		if (oldp)
			free(oldp);
		return NULL;
	}
	if (sz > SIZE_MAX / 2)
		return NULL;
	lock(&rblock);
	n.buf = oldp;
	res = RB_FIND(alloc_tree, &at, &n);
	if (res) {
		if (res->sz == sz) {
			unlock(&rblock);
			st.nr_realloc++;
			return res->buf;
		}
		if (res->sz > sz) {
			size_t diff = res->sz - sz;
			/* We cannot ensure the alignment of the free
			 * block if its size is less than ALIGN - so just
			 * don't split the node in that case */
			if (diff < ALIGN) {
				unlock(&rblock);
				st.nr_realloc++;
				return res->buf;
			}
			newfn = alloc_object(sizeof(*newfn));
			if (!newfn) {
				unlock(&rblock);
				return NULL;
			}
			res->sz = sz;
			newfn->buf = res->buf + sz;
			newfn->sz = diff;
			align_node(newfn);
			RB_INSERT(free_tree, &ft, newfn);
			unlock(&rblock);
			st.nr_realloc++;
			st.nr_shrink_realloc++;
			return res->buf;
		}
		oldan = res;
		/* Lookup in the free tree for a block greater
		 * than or equal to `sz' bytes */
		n.sz = sz;
		res = RB_NFIND(free_tree, &ft, &n);
		if (!res) {
			/* No available free block, create a new block
			 * and add it to the alloc tree */
			newan = alloc_object(sizeof(*newan));
			if (!newan) {
				unlock(&rblock);
				return NULL;
			}
			newan->buf = mmap_aligned(ALIGN, sz);
			if (!newan->buf) {
				unlock(&rblock);
				return NULL;
			}
			newan->sz = sz;
			RB_INSERT(alloc_tree, &at, newan);
			st.nr_alloc_new++;
		} else {
			/* Grab the block from the free tree instead */
			newan = RB_REMOVE(free_tree, &ft, res);
			RB_INSERT(alloc_tree, &at, newan);
			st.nr_alloc_free++;
		}
		/* Copy over the contents from `oldp' to the
		 * new memory block */
		memcpy(newan->buf, oldan->buf,
		       sz < oldan->sz ? sz : oldan->sz);
		/* Return `oldp' to the free tree */
		fn = RB_REMOVE(alloc_tree, &at, oldan);
		RB_INSERT(free_tree, &ft, fn);
		unlock(&rblock);
		st.nr_realloc++;
		return newan->buf;
	}
	unlock(&rblock);
	return NULL;
}

void *
calloc(size_t nmemb, size_t sz)
{
	void *p;

	if (lemon_init())
		return NULL;
	p = malloc(nmemb * sz);
	if (!p)
		return NULL;
	memset(p, 0, nmemb * sz);
	return p;
}

void
free(void *p)
{
	struct node n, *fn, *res;

	if (lemon_init())
		return;
	if (!p)
		return;
	lock(&rblock);
	n.buf = p;
	res = RB_FIND(alloc_tree, &at, &n);
	if (res) {
		fn = RB_REMOVE(alloc_tree, &at, res);
		RB_INSERT(free_tree, &ft, fn);
		st.nr_free++;
	} else {
		st.nr_invalid_free++;
		if (strict) {
			writelog(STDERR_FILENO, "*** Freeing invalid pointer: %p, aborting ***\n", p);
			_Exit(1);
		}
	}
	unlock(&rblock);
}

void
cfree(void *p)
{
	return free(p);
}

void *
memalign(size_t align, size_t sz)
{
	struct node *an;
	void *p;

	if (lemon_init())
		return NULL;
	if (((align - 1) & align))
		return NULL;
	if (align < sizeof(void *))
		return NULL;
	if (align > SIZE_MAX - align)
		return NULL;
	/* Just allocate a new block, we don't care to look
	 * for a block in the free tree as it might not be properly
	 * aligned.  The previous implementation could cope with
	 * that but it was sort of hackish.  There are few calls to
	 * posix_memalign() in most cases, so the overhead should
	 * not really matter. */
	an = alloc_object(sizeof(*an));
	if (!an)
		return NULL;
	p = mmap_aligned(align, sz);
	if (!p)
		return NULL;
	an->buf = p;
	an->sz = sz;
	lock(&rblock);
	RB_INSERT(alloc_tree, &at, an);
	unlock(&rblock);
	st.nr_alloc_new++;
	return p;
}

void *
aligned_alloc(size_t align, size_t sz)
{
	if (sz % align)
		return NULL;
	return memalign(align, sz);
}

void *
valloc(size_t sz)
{
	return memalign(sysconf(_SC_PAGESIZE), sz);
}

void *
pvalloc(size_t sz)
{
	long pagesize = sysconf(_SC_PAGESIZE);
	sz = pagesize * ((sz + pagesize - 1) / pagesize);
	return valloc(sz);
}

int
posix_memalign(void **memptr, size_t align, size_t sz)
{
	*memptr = NULL;
	if (((align - 1) & align))
		return EINVAL;
	if (align < sizeof(void *))
		return EINVAL;
	if (sz > SIZE_MAX - align)
		return ENOMEM;
	*memptr = memalign(align, sz);
	if (!*memptr)
		return ENOMEM;
	return 0;
}

size_t
malloc_usable_size(void *p)
{
	struct node n, *res;

	if (lemon_init())
		return 0;
	if (!p)
		return 0;
	lock(&rblock);
	n.buf = p;
	res = RB_FIND(alloc_tree, &at, &n);
	if (res) {
		unlock(&rblock);
		return res->sz;
	}
	unlock(&rblock);
	return 0;
}

size_t
malloc_size(void *p)
{
	return malloc_usable_size(p);
}

static void
writelog(int fd, const char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZ];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	write(fd, buf, strlen(buf));
}

static void
dumpstats(void)
{
	int fd;
	struct node *n;

	fd = open("lemoncake.out", O_WRONLY | O_CREAT | O_APPEND, 0644);
	if (fd < 0)
		return;
	writelog(fd, "*** Lemoncake stats [process: %s pid: %lu] ***\n",
		 getenv("_") ? getenv("_") : "<unknown>", (unsigned long)getpid());
	writelog(fd, "sbrk calls: %lu\n", st.nr_sbrk);
	writelog(fd, "mmap calls: %lu\n", st.nr_mmap);
	writelog(fd, "malloc calls: %lu\n", st.nr_malloc);
	writelog(fd, "realloc calls: %lu\n", st.nr_realloc);
	writelog(fd, "shrinked realloc calls (no memcpy() required): %lu\n",
		 st.nr_shrink_realloc);
	writelog(fd, "free calls: %lu\n", st.nr_free);
	writelog(fd, "new allocations: %lu\n", st.nr_alloc_new);
	writelog(fd, "allocations from the free tree: %lu\n",
		 st.nr_alloc_free);
	writelog(fd, "invalid free calls: %lu\n", st.nr_invalid_free);
	RB_FOREACH(n, alloc_tree, &at)
		st.nr_at_nodes++;
	RB_FOREACH(n, free_tree, &ft)
		st.nr_ft_nodes++;
	writelog(fd, "nodes in the allocation tree: %lu\n",
		 st.nr_at_nodes);
	writelog(fd, "nodes in the free tree: %lu\n",
		 st.nr_ft_nodes);
	close(fd);
}
